package de.tudresden.inf.st.mquat.eris;

import de.tudresden.inf.st.mquat.eris.coupling.*;
import de.tudresden.inf.st.mquat.jastadd.model.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;


public class ErisMQuATBuilder {

  private static final Logger logger = LogManager.getLogger(ErisMQuATBuilder.class);

  static Root createModel(ERISModel erisModel) {

    Root root = new Root();

    // set the objective property runtime
    Property runtime = new Property();
    runtime.setName(new Name("runtime"));
    runtime.setUnit("ms");

    // set hardware model
    HardwareModel hardwareModel = new HardwareModel();

    ResourceType accessPath = new ResourceType();
    accessPath.setName(new Name("AccessPath"));
    accessPath.setContainer(true);
    Property dynamic = new Property(new Name("dynamic"), "flag");
    Property attributeCount = new Property(new Name("attributeCount"), "#");
    Property attributeNull = new Property(new Name("attributeNull"), "flag");
    Property attributeUndefined = new Property(new Name("attributeUndefined"), "flag");
    Property dynamicAttributeSize = new Property(new Name("dynamicAttributeSize"), "flag");
    Property indexAttributeCount = new Property(new Name("indexAttributeCount"), "flag");
    Property indexOrderPreserving = new Property(new Name("indexOrderPreserving"), "flag");
    Property direct = new Property(new Name("isDirect"), "flag");
    Property index = new Property(new Name("isIndex"), "flag");
    accessPath.addProperty(dynamic);
    accessPath.addProperty(attributeCount);
    accessPath.addProperty(attributeNull);
    accessPath.addProperty(attributeUndefined);
    accessPath.addProperty(dynamicAttributeSize);
    accessPath.addProperty(indexAttributeCount);
    accessPath.addProperty(indexOrderPreserving);
    accessPath.addProperty(direct);
    accessPath.addProperty(index);

    // add energy benchmark properties
    Map<Integer, Property> insertProperties = new HashMap<>();
    Map<Integer, Property> linkProperties = new HashMap<>();
    Map<Integer, Property> lookupProperties = new HashMap<>();
    Map<Integer, Property> scanProperties = new HashMap<>();
    for (ERISBenchmarkResult result : erisModel.getAccessPaths().iterator().next().getResultList()) {
      int size = result.getDataSize();
      insertProperties.put(size, new Property(new Name("insertTime" + size), "ms"));
      accessPath.addProperty(insertProperties.get(size));
      linkProperties.put(size, new Property(new Name("linkTime" + size), "ms"));
      accessPath.addProperty(linkProperties.get(size));
      lookupProperties.put(size, new Property(new Name("lookupTime" + size), "ms"));
      accessPath.addProperty(lookupProperties.get(size));
      scanProperties.put(size, new Property(new Name("scanTime" + size), "ms"));
      accessPath.addProperty(scanProperties.get(size));
    }

    hardwareModel.addResourceType(accessPath);

    // add a resource for every access path
    for (ERISAccessPath erisAccessPath : erisModel.getAccessPaths()) {

      MultiResource multiResource = new MultiResource();

      // add properties
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(dynamic.createRef(), new LiteralExpression(erisAccessPath.isDynamic() ? 1 : 0))
      );
      LiteralExpression attributeCountValue;
      switch (erisAccessPath.getAttributeCount()) {
        case "dynamic":
        case "any":
          attributeCountValue = new LiteralExpression(Double.POSITIVE_INFINITY);
          break;
        default:
          attributeCountValue = new LiteralExpression(Double.valueOf(erisAccessPath.getAttributeCount()));
      }
      multiResource.addCurrentResourceValue(new CurrentResourceValue(attributeCount.createRef(), attributeCountValue));
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(attributeNull.createRef(), new LiteralExpression(erisAccessPath.getAttributeNull()))
      );
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(attributeUndefined.createRef(), new LiteralExpression(erisAccessPath.getAttributeUndefined()))
      );
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(dynamicAttributeSize.createRef(), new LiteralExpression(erisAccessPath.getDynamicAttributeSize()))
      );
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(indexAttributeCount.createRef(), new LiteralExpression(erisAccessPath.getIndexAttributeCount()))
      );
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(indexOrderPreserving.createRef(), new LiteralExpression(erisAccessPath.getIndexOrderPreserving()))
      );
      boolean isDirect = erisAccessPath.getAddressingMode().equals("direct");
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(direct.createRef(), new LiteralExpression(isDirect ? 1D : 0D))
      );
      boolean isIndex = erisAccessPath.getAddressingMode().equals("index");
      multiResource.addCurrentResourceValue(
          new CurrentResourceValue(index.createRef(), new LiteralExpression(isIndex ? 1D : 0D))
      );

      // add all the attribute stat values
      for (ERISBenchmarkResult benchmarkResult : erisAccessPath.getResultList()) {
        Property insertProperty = insertProperties.get(benchmarkResult.getDataSize());
        multiResource.addCurrentResourceValue(
            new CurrentResourceValue(insertProperty.createRef(), new LiteralExpression(benchmarkResult.getInsert()))
        );
        Property linkProperty = linkProperties.get(benchmarkResult.getDataSize());
        multiResource.addCurrentResourceValue(
            new CurrentResourceValue(linkProperty.createRef(), new LiteralExpression(benchmarkResult.getLink()))
        );
        Property lookupProperty = lookupProperties.get(benchmarkResult.getDataSize());
        multiResource.addCurrentResourceValue(
            new CurrentResourceValue(lookupProperty.createRef(), new LiteralExpression(benchmarkResult.getLookup()))
        );
        Property scanProperty = scanProperties.get(benchmarkResult.getDataSize());
        multiResource.addCurrentResourceValue(
            new CurrentResourceValue(scanProperty.createRef(), new LiteralExpression(benchmarkResult.getScan()))
        );
      }

      // add resources
      if (erisAccessPath.isDynamic()) {
        multiResource.setName(
            new Name("dynamic_" +
                erisAccessPath.getName() +
                "_" + erisAccessPath.getId() +
                "_" + erisAccessPath.getAddressingMode()));
      } else {
        multiResource.setName(
            new Name(erisAccessPath.getName() +
                "_" + erisAccessPath.getId() +
                "_" + erisAccessPath.getAddressingMode()));
      }

      multiResource.setAmount(1000000); // TODO determine how many resources are required
      multiResource.setType(accessPath.createRef());
      hardwareModel.addResource(multiResource);
    }


    // set software model
    SoftwareModel softwareModel = new SoftwareModel();


    // for each container, living partition, and attribute, create a software component and a request
    for (ERISContainer erisContainer : erisModel.getContainers()) {
      for (ERISAttribute attribute : erisContainer.getAttributes()) {

        // ignore linkage attributes
        if (attribute.isLinkage()) continue;


        for (ERISLivingPartition erisLivingPartition : erisContainer.getPartitions()) {

          if (erisLivingPartition.isIgnore()) {
            continue;
          }

          // ignore attribute if all stats are zero
          ERISAttributeStatistics statistics = erisLivingPartition.getAttributeStatistics(attribute);

          // if there are no stats available, just move on to the next LP
          if (statistics == null) {
            continue;
          }

          if (statistics.getInserts() == 0
              && statistics.getLookups() == 0
              && statistics.getLinks() == 0
              && statistics.getScans() == 0) {
            logger.debug("counters are all zero for attribute {}", attribute.getId());
            continue;
          }

          if (erisLivingPartition.isAdaptationScheduled()) {
            logger.debug("Skipping living partition {}, because an adaptation is already scheduled", erisLivingPartition.getLocalId());
            continue;
          }

          Component directOrIndirectComponent = new Component();
          directOrIndirectComponent.setName(new Name("directOrIndirect_" + createName(erisContainer, erisLivingPartition, attribute)));
          Implementation directOrIndirectImplementation = new Implementation();
          directOrIndirectImplementation.setName(new Name("directOrIndirectAccessPathSelector"));

          // create the requirement for an access path
          Instance directOrIndirectAccessPathInstance = new Instance(new Name("directOrIndirectAccessPath_" + attribute.getName()));
          ResourceRequirement directOrIndirectResourceRequirement = new ResourceRequirement();
          directOrIndirectResourceRequirement.addInstance(directOrIndirectAccessPathInstance);
          directOrIndirectResourceRequirement.setResourceTypeRef(accessPath.createRef());
          directOrIndirectImplementation.setResourceRequirement(directOrIndirectResourceRequirement);

          // ensure access path properties
          // on a per-attribute access path configuration, only the null and undefined checks have to be done
          if (!attribute.isNotNull()) {
            directOrIndirectImplementation.addClause(new Clause(
                ClauseType.REQUIRING,
                new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), attributeNull.createRef()),
                ClauseComparator.EQ,
                new LiteralExpression(1D)
            ));
          }
          if (!attribute.isNotUndefined()) {
            directOrIndirectImplementation.addClause(new Clause(
                ClauseType.REQUIRING,
                new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), attributeUndefined.createRef()),
                ClauseComparator.EQ,
                new LiteralExpression(1D)
            ));
          }
          // if the attribute has a dynamic length, require a dynamic attribute size
          if (attribute.getTypeLength() == 0) {
            directOrIndirectImplementation.addClause(new Clause(
                ClauseType.REQUIRING,
                new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), dynamicAttributeSize.createRef()),
                ClauseComparator.EQ,
                new LiteralExpression(1D)
            ));
          }

          // create the provisions

          // score is:
          // score = this.lookupTime*this.numLookups + this.insertTime*this.numInserts


          int num_lookups = statistics.getLookups();
          int num_inserts = statistics.getInserts();
          int num_scans = statistics.getScans();
          int num_links = statistics.getLinks();
          int att_count = statistics.getCount();

          // get the correct size
          int lookupsize = getLookupSize(lookupProperties.keySet(), att_count);

          {
            MultExpression loopups = new MultExpression(new LiteralExpression(num_lookups), new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), lookupProperties.get(lookupsize).createRef()));
            MultExpression inserts = new MultExpression(new LiteralExpression(num_inserts), new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), insertProperties.get(lookupsize).createRef()));
            MultExpression scans = new MultExpression(new MultExpression(new LiteralExpression(num_scans), new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), scanProperties.get(lookupsize).createRef())), new LiteralExpression(att_count));
            MultExpression links = new MultExpression(new LiteralExpression(num_links), new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), linkProperties.get(lookupsize).createRef()));

            Expression directOrIndirectExpression = new AddExpression(loopups, new AddExpression(inserts, new AddExpression(scans, links)));
            directOrIndirectImplementation.addClause(new Clause(ClauseType.PROVIDING, new SoftwareDesignator(new Opt<>(), runtime.createRef()), ClauseComparator.EQ, directOrIndirectExpression));
          }

          // add an additional requirement to only allow direct or indirect access paths for the direct implementation
          directOrIndirectImplementation.addClause(new Clause(
              ClauseType.REQUIRING,
              new PropertyResourceDesignator(directOrIndirectAccessPathInstance.createRef(), index.createRef()),
              ClauseComparator.EQ,
              new LiteralExpression(0D)
          ));


          directOrIndirectComponent.addImplementation(directOrIndirectImplementation);

          // no lookups means no index
          if (attribute.isAllowIndex() && num_lookups > 0) {
            Component indexComponent = new Component();
            indexComponent.setName(new Name("index_" + createName(erisContainer, erisLivingPartition, attribute)));
            Implementation indexImplementation = new Implementation();
            indexImplementation.setName(new Name("indexAccessPathSelector"));

            // create the requirement for an access path
            Instance indexAccessPathInstance = new Instance(new Name("indexAccessPath_" + attribute.getName()));
            ResourceRequirement indexResourceRequirement = new ResourceRequirement();
            indexResourceRequirement.addInstance(indexAccessPathInstance);
            indexResourceRequirement.setResourceTypeRef(accessPath.createRef());
            indexImplementation.setResourceRequirement(indexResourceRequirement);

            if (!attribute.isNotNull()) {
              indexImplementation.addClause(new Clause(
                  ClauseType.REQUIRING,
                  new PropertyResourceDesignator(indexAccessPathInstance.createRef(), attributeNull.createRef()),
                  ClauseComparator.EQ,
                  new LiteralExpression(1D)
              ));
            }
            if (!attribute.isNotUndefined()) {
              indexImplementation.addClause(new Clause(
                  ClauseType.REQUIRING,
                  new PropertyResourceDesignator(indexAccessPathInstance.createRef(), attributeUndefined.createRef()),
                  ClauseComparator.EQ,
                  new LiteralExpression(1D)
              ));
            }
            // if the attribute has a dynamic length, require a dynamic attribute size
            if (attribute.getTypeLength() == 0) {
              indexImplementation.addClause(new Clause(
                  ClauseType.REQUIRING,
                  new PropertyResourceDesignator(indexAccessPathInstance.createRef(), dynamicAttributeSize.createRef()),
                  ClauseComparator.EQ,
                  new LiteralExpression(1D)
              ));
            }

            {
              MultExpression lookups = new MultExpression(new LiteralExpression(num_lookups), new PropertyResourceDesignator(indexAccessPathInstance.createRef(), lookupProperties.get(lookupsize).createRef()));
              MultExpression inserts = new MultExpression(new LiteralExpression(num_inserts), new PropertyResourceDesignator(indexAccessPathInstance.createRef(), insertProperties.get(lookupsize).createRef()));
              Expression indexExpression = new AddExpression(lookups, inserts);
              indexImplementation.addClause(new Clause(ClauseType.PROVIDING, new SoftwareDesignator(new Opt<>(), runtime.createRef()), ClauseComparator.EQ, indexExpression));
            }

            // add an additional requirement to only index access paths for the direct implementation
            indexImplementation.addClause(new Clause(
                ClauseType.REQUIRING,
                new PropertyResourceDesignator(indexAccessPathInstance.createRef(), index.createRef()),
                ClauseComparator.EQ,
                new LiteralExpression(1D)
            ));

            indexComponent.addImplementation(indexImplementation);

            softwareModel.addComponent(indexComponent);

            // create the index request
            Request indexRequest = new Request();
            indexRequest.setTarget(indexComponent.createRef());
            indexRequest.setName(new Name("index_request_" + createName(erisContainer, erisLivingPartition, attribute)));
            root.addRequest(indexRequest);
          }

          softwareModel.addComponent(directOrIndirectComponent);

          // create the direct or indirect request
          Request directOrIndirectRequest = new Request();
          directOrIndirectRequest.setTarget(directOrIndirectComponent.createRef());
          directOrIndirectRequest.setName(new Name("request_" + createName(erisContainer, erisLivingPartition, attribute)));
          root.addRequest(directOrIndirectRequest);
        }
      }
    }
    // set objective
    Objective objective = new Objective();
    objective.setPropertyRef(runtime.createRef());
    objective.setAgg(PropertyAggregation.SUM);

    root.setSoftwareModel(softwareModel);
    root.setHardwareModel(hardwareModel);
    root.setObjective(objective);

    return root;
  }

  private static int getLookupSize(Collection<Integer> lookupSizes, int att_count) {
    // get the correct size
    int lookupsize = 0;
    int dist = Integer.MAX_VALUE;
    for (Integer i : lookupSizes) {
      if (Math.abs(att_count - i) < dist) {
        dist = Math.abs(att_count - i);
        lookupsize = i;
      }
    }
    return lookupsize;
  }

  private static String createName(ERISContainer container, ERISLivingPartition partition, ERISAttribute attribute) {
    // TODO use ID instead of name for attribute if possible
    return container.getId() + "_" + partition.getNodeId() + "_" + partition.getLocalId() + "_" + attribute.getId() + "_" + attribute.getName();
  }

}
