package de.tudresden.inf.st.mquat.eris.coupling;

import de.tudresden.inf.st.mquat.jastadd.model.Assignment;
import de.tudresden.inf.st.mquat.jastadd.model.Solution;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;

public class ERISConnector {

  private static final Logger logger = LogManager.getLogger(ERISConnector.class);

  private final String ERIS_URL = "127.0.0.1:5189";
  private String sessionID;

  private ERISModel erisModel = new ERISModel();

  public ERISConnector() {
    // curl --data "login=jo&password=jo" http://127.0.0.1:5189/session
    try {
      URI uri = new URIBuilder()
          .setHost(ERIS_URL)
          .setScheme("http")
          .setPath("/session")
          .setParameter("login", "b01")
          .setParameter("password", "b01")
          .build();

      JSONObject jsonObject = getSessionID(uri);

      sessionID = (String) jsonObject.get("id");

      logger.info("Created Session with ID {}", sessionID);
    } catch (URISyntaxException | JSONException e) {
      logger.error("Creation of ERIS connector failed: {}", e.getMessage());
      throw new RuntimeException(e);
    }
  }

  public void read() {
    // curl http://127.0.0.1:5189/swctrl/accesspaths?id=
    try {
      URI uri = new URIBuilder()
          .setHost(ERIS_URL)
          .setScheme("http")
          .setPath("swctrl/accesspaths")
          .setParameter("id", sessionID)
          .build();

      String content = getContent(uri);
      //String content = new String(Files.readAllBytes(new File("jastadd-mquat-base/src/main/resources/accesspaths.json").toPath()));

      JSONArray accesspathList = new JSONArray(content);

      for (int i = 0; i < accesspathList.length(); i++) {
        erisModel.addAccessPath(new ERISAccessPath(accesspathList.getJSONObject(i)));
      }
    } catch (URISyntaxException | JSONException e) {
      logger.error("Reading of access paths failed: {}", e.getMessage());
      throw new RuntimeException(e);
    }

    // curl http://127.0.0.1:5189/swctrl/containers?id=<>

    try {
      URI uri = new URIBuilder()
          .setHost(ERIS_URL)
          .setScheme("http")
          .setPath("swctrl/containers")
          .setParameter("id", sessionID)
          .build();

      String content = getContent(uri);
      JSONArray containerList = new JSONArray(content);

      for (int i = 0; i < containerList.length(); i++)
        erisModel.addContainer(new ERISContainer(containerList.getJSONObject(i), erisModel));

      // curl http://127.0.0.1:5189/swctrl/container/<cID>/?id=
      for (ERISContainer ec : erisModel.getContainers()) {
        URI uri2 = new URIBuilder()
            .setHost(ERIS_URL)
            .setScheme("http")
            .setPath("swctrl/container/" + ec.getId())
            .setParameter("id", sessionID)
            .build();

        content = getContent(uri2);
        JSONObject containerInfo = new JSONObject(content);

        JSONArray attributes = (JSONArray) containerInfo.get("attributes");
        for (int i = 0; i < attributes.length(); i++)
          ec.addAttribute(new ERISAttribute(attributes.getJSONObject(i), ec));

        JSONArray lps = (JSONArray) containerInfo.get("lps");
        for (int i = 0; i < lps.length(); i++)
          ec.addPartition(new ERISLivingPartition(lps.getJSONObject(i), ec));
      }

      // curl http://127.0.0.1:5189/swctrl/lp/<cID>/<nodeID>/<localID>?id=

      List<ERISLivingPartition> livingPartitionsToIgnore = new ArrayList<>();

      for (ERISContainer ec : erisModel.getContainers()) {
        for (ERISLivingPartition lp : ec.getPartitions()) {
          URI uri3 = new URIBuilder()
              .setHost(ERIS_URL)
              .setScheme("http")
              .setPath("swctrl/lp/" + ec.getId() + "/" + lp.getNodeId() + "/" + lp.getLocalId())
              .setParameter("id", sessionID)
              .build();
          content = getContent(uri3);
          JSONObject lpInfo = new JSONObject(content);
          JSONArray configurations = lpInfo.getJSONArray("configuration");

          if (configurations.length() == 1) {
            String errorMessage = configurations.getJSONObject(0).optString("error");
            if (errorMessage != null) {
              logger.info("Skipping partition {}/{}/{} because '{}'", lp.getContainer(), lp.getNodeId(), lp.getLocalId(), errorMessage);
              livingPartitionsToIgnore.add(lp);
              lp.setIgnore(true);
              continue;
            }
          }

          for (int i = 0; i < configurations.length(); i++)
            lp.getConfigurations().
                add(new ERISLPConfiguration(configurations.getJSONObject(i), lp));

          JSONObject stats = lpInfo.getJSONObject("statistics");
          lp.setStatistics(new ERISLPStatistics(stats));

          JSONArray attrStats = stats.getJSONArray("attributes");
          for (int i = 0; i < attrStats.length(); i++) {
            ERISAttribute a = ec.getAttribute(attrStats.getJSONObject(i).getInt("attributeId"));
            lp.addAttributeStatistics(new ERISAttributeStatistics(attrStats.getJSONObject(i), a, lp));
          }
        }
      }
    } catch (URISyntaxException | JSONException e) {
      logger.error("Getting containers failed: {}", e.getMessage());
      throw new RuntimeException(e);
    }
  }


  private JSONObject getSessionID(URI uri) {
    try {
      HttpPost httpPost = new HttpPost(uri);
      logger.info("Sending {}", httpPost.toString());
      CloseableHttpClient client = HttpClients.createDefault();

      CloseableHttpResponse response = client.execute(httpPost);
      Scanner s = new Scanner(response.getEntity().getContent()).useDelimiter("\\A");
      String result = s.hasNext() ? s.next() : "";
      client.close();
      response.close();

      return new JSONObject(result);
    } catch (UnsupportedOperationException | IOException | JSONException e) {
      logger.error("Getting session id from {} failed: {}", uri.toString(), e.getMessage());
      throw new RuntimeException(e);
    }
  }

  private String getContent(URI uri) {
    try {
      HttpGet httpGet = new HttpGet(uri);
      logger.info("Sending {}", httpGet.toString());
      CloseableHttpClient client = HttpClients.createDefault();

      CloseableHttpResponse response = client.execute(httpGet);
      Scanner s = new Scanner(response.getEntity().getContent()).useDelimiter("\\A");
      String result = s.hasNext() ? s.next() : "";

      client.close();
      response.close();

      return result;

    } catch (UnsupportedOperationException | IOException e) {
      logger.error("Reading from {} failed: {}", uri.toString(), e.getMessage());
      throw new RuntimeException(e);
    }
  }

  public void write(Solution solution) {
    // maps mapping from name ({container}_{node}_{lp}_{attribute#}_{attributename}) to the assignment
    // the maps are for the two request types
    Map<ERISLivingPartition, Map<ERISAttribute, ERISLPConfiguration>> directOrIndirectAssignments = new HashMap<>();
    Map<ERISLivingPartition, Map<ERISAttribute, ERISLPConfiguration>> indexAssignments = new HashMap<>();

    for (ERISContainer container : erisModel.getContainers()) {
      for (ERISLivingPartition livingPartition : container.getPartitions()) {
        directOrIndirectAssignments.put(livingPartition, new HashMap<>());
        indexAssignments.put(livingPartition, new HashMap<>());
      }
    }

    // put the assignments in the corresponding maps
    for (Assignment assignment : solution.getAssignmentList()) {
      String name = assignment.getRequest().name();
      boolean index = false;
      if (name.startsWith("index_")) {
        index = true;
        name = name.substring("index_request_".length());
      } else {
        name = name.substring("request_".length());
      }
      String[] nameArray = name.split("_");
      ERISContainer container = erisModel.getContainer(Integer.valueOf(nameArray[0]));
      ERISLivingPartition livingPartition = container.getPartition(Integer.valueOf(nameArray[1]), Integer.valueOf(nameArray[2]));
      ERISAttribute attribute = container.getAttribute(Integer.valueOf(nameArray[3]));

      String[] accessPathName = assignment.getResourceMapping().getResource().name().split("_");

      ERISAccessPath accessPath = erisModel.getAccessPath(Integer.valueOf(accessPathName[1]), accessPathName[2]);
      ERISLPConfiguration configuration = new ERISLPConfiguration(accessPath, Collections.singletonList(attribute), new ArrayList<>(), livingPartition, assignment.computeObjective());

      if (index) {
        indexAssignments.get(livingPartition).put(attribute, configuration);
      } else {
        directOrIndirectAssignments.get(livingPartition).put(attribute, configuration);
      }

    }

//    logger.debug("Overview over found new configurations");
//    for (ERISContainer container : erisModel.getContainers()) {
//      logger.debug("  Container {}:", container);
//      for (ERISLivingPartition partition : container.getPartitions()) {
//        logger.debug("    Partition {}:", partition);
//        if (directOrIndirectAssignments.get(partition).isEmpty()) {
//          logger.debug("      no direct or indirect assignments!");
//        } else {
//          logger.debug("      direct or indirect assignments");
//          for (ERISAttribute attribute : directOrIndirectAssignments.get(partition).keySet()) {
//            ERISLPConfiguration configuration = directOrIndirectAssignments.get(partition).get(attribute);
//            logger.debug("        put {} to get an objective of {}", configuration, configuration.getEstimatedCost().orElse(-1D));
//          }
//        }
//        if (indexAssignments.get(partition).isEmpty()) {
//          logger.debug("      no index assignments!");
//        } else {
//          logger.debug("      index assignments");
//          for (ERISAttribute attribute : indexAssignments.get(partition).keySet()) {
//            ERISLPConfiguration configuration = indexAssignments.get(partition).get(attribute);
//            logger.debug("        put {} to get an objective of {}", configuration, configuration.getEstimatedCost().orElse(-1D));
//          }
//        }
//      }
//    }

    // for each living partition, check, if the assignment has changed and if an index should be sent
    logger.debug("Overview over found new configurations");
    for (ERISContainer container : erisModel.getContainers()) {
      logger.debug("  Container {}:", container);
      for (ERISLivingPartition livingPartition : container.getPartitions()) {
        logger.debug("    Living Partition {}:", livingPartition);
        // variable to check if the config has changed for any attribute in the living partition. If so, the whole new
        // config has to be sent!
        boolean configHasChanged = false;
        List<ERISLPConfiguration> configsToUse = new ArrayList<>();

        Map<ERISAttribute, ERISLPConfiguration> currentDOrIConfigs = directOrIndirectAssignments.get(livingPartition);
        Map<ERISAttribute, ERISLPConfiguration> currentIndexConfigs = indexAssignments.get(livingPartition);

        for (ERISAttribute attribute : container.getAttributes()) {

          if (currentDOrIConfigs.containsKey(attribute)) {
            // get the new config
            ERISLPConfiguration newDOrOConfig = currentDOrIConfigs.get(attribute);

            ERISAttributeStatistics attributeStatistics = livingPartition.getAttributeStatistics(attribute);
            logger.debug("      attribute {}:", attribute.getName());
            logger.debug("        stats: lookups: {} inserts: {} scans: {} links: {} count: {}",
                attributeStatistics.getLookups(),
                attributeStatistics.getInserts(),
                attributeStatistics.getScans(),
                attributeStatistics.getLinks(),
                attributeStatistics.getCount());

            // get the old config for the same attribute (if there is one)
            Optional<ERISLPConfiguration> oldConfig = livingPartition.getConfigForAttribute(attribute, false);


            boolean localChange;
            if (oldConfig.isPresent()) {
              localChange = oldConfig.get().getAccessPath().getUId().equals(newDOrOConfig.getAccessPath().getUId());
              if (localChange) {
                configsToUse.add(newDOrOConfig);
                configHasChanged = true;
              } else oldConfig.ifPresent(configsToUse::add);
            } else {
              logger.debug("        NO old direct or indirect access path for attribute {} was found!", attribute.getName());
              configsToUse.add(newDOrOConfig);
              localChange = true;
              configHasChanged = true;
            }

            logger.debug("        put {} to get an objective of {}{}", newDOrOConfig, newDOrOConfig.getEstimatedCost().orElse(Double.MAX_VALUE), localChange ? ("; this is a change") : "");


            // check if an index should be used

            // first prerequisite: there has to be a proposed index
            if (currentIndexConfigs.containsKey(attribute)) {

              // get the proposed index
              ERISLPConfiguration newIndexConfig = currentIndexConfigs.get(attribute);

              // compute the missing component of the score for the index config

              // get the lookup time
              int count = livingPartition.getAttributeStatistics(attribute).getCount();
              int lookupCount = livingPartition.getAttributeStatistics(attribute).getLookups();
              double lookupTime = newDOrOConfig.getAccessPath().getLookupTime(count);
              double indexScore = newIndexConfig.getEstimatedCost().orElse(Double.MAX_VALUE);
              double lookupScoreWithoutIndex = lookupTime * lookupCount;

              if (indexScore < lookupScoreWithoutIndex) {
                // the index is good for us (if we ignore the creation cost!)

                // get the old config for the same attribute (if there is one)
                Optional<ERISLPConfiguration> oldIndexConfig = livingPartition.getConfigForAttribute(attribute, true);

                boolean localIndexChange;
                if (oldIndexConfig.isPresent()) {
                  localIndexChange = oldIndexConfig.get().getAccessPath().getUId().equals(newIndexConfig.getAccessPath().getUId());
                  if (localIndexChange) {
                    configsToUse.add(newIndexConfig);
                    configHasChanged = true;
                  } else oldIndexConfig.ifPresent(configsToUse::add);
                } else {
                  logger.debug("        NO old index access path for attribute {} was found!", attribute.getName());
                  configsToUse.add(newIndexConfig);
                  localIndexChange = true;
                  configHasChanged = true;
                }
                logger.debug("        create an index {} to get an (index-) objective of {} (compared to {} without an index){}", newIndexConfig, indexScore, lookupScoreWithoutIndex, localIndexChange ? ("; this is a change") : "");

                // if an index is created, only CREATE the direct/indirect access path if there are scans happening!
                // if, however, it already exists, keep it
                if (localIndexChange && localChange) {
                  if (attributeStatistics.getScans() == 0) {
                    if (!configsToUse.remove(newDOrOConfig)) {
                      logger.error("Trying to remove config from new access path that is not in it: {}", newDOrOConfig);
                    } else {
                      logger.debug("        NOT creating direct/indirect access path {}, because we also have an index, but {} scans.", newDOrOConfig, attributeStatistics.getScans());
                    }
                  }
                }

              } else {
                logger.debug("        do NOT create an index {} to get an (index-) objective of {} (compared to {} without an index)", newIndexConfig, indexScore, lookupScoreWithoutIndex);
              }
            } else {
              // TODO this is our guess, but it seems reasonable to not get rid of an index if it causes no harm
              // if there are no inserts, keep a potentially existing index
              if (attributeStatistics.getInserts() == 0) {
                // see if there is an index for the current attribute
                livingPartition.getConfigForAttribute(attribute, true).ifPresent(configsToUse::add);
              }
            }
          }
        }
        // figure out if the config has changed by comparing the new and old config by access path and attributes

        // check if all new configs are in oldconfig
        configHasChanged = false;
        outer:
        for (ERISLPConfiguration newConfig : configsToUse) {
          for (ERISLPConfiguration oldConfig : livingPartition.getConfigurations()) {
            if (oldConfig.getAccessPath() == newConfig.getAccessPath() && oldConfig.getAttributesAndKeys().containsAll(newConfig.getAttributesAndKeys())) {
              continue outer;
            }
          }
          configHasChanged = true;
          break;
        }
        logger.debug("          Config change detector: Additional attributes have been added: {}", configHasChanged);
        if (!configHasChanged) {
          outer:
          for (ERISLPConfiguration oldConfig : livingPartition.getConfigurations()) {
            // skip default ap
            if (oldConfig.getAttributesAndKeys().isEmpty()) continue;

            // if the old config contains only linkage attributes, continue
            boolean linkageOnly = true;
            for (ERISAttribute attribute : oldConfig.getAttributesAndKeys()) {
              if (!attribute.isLinkage()) {
                linkageOnly = false;
              }
            }
            if (linkageOnly) continue;

            for (ERISLPConfiguration newConfig : configsToUse) {
              if (oldConfig.getAccessPath() == newConfig.getAccessPath() && oldConfig.getAttributesAndKeys().containsAll(newConfig.getAttributesAndKeys())) {
                continue outer;
              }
            }
            configHasChanged = true;
            break;
          }
          logger.debug("          Config change detector: Attributes have been removed: {}", configHasChanged);
        }


        logger.debug("        -> the configuration change detector says config has changed for living partition {}: {}", livingPartition, configHasChanged);

        logger.debug("          old config");
        for (ERISLPConfiguration oldConfig : livingPartition.getConfigurations()) {
          logger.debug("          {}", oldConfig);
        }
        logger.debug("          new config");
        for (ERISLPConfiguration newConfig : configsToUse) {
          logger.debug("          {}", newConfig);
        }


        if (configHasChanged) {
          // write back changed config

          /*
           * POST /swctrl/lp/<container id>/<node id>/<local id>
           *
           * Body: {storageConfigurationGroups: [<group>, <group>]}
           * <group>: {attributes: [<attr>, <attr>], newAPId: <id>, newAPisDynamic: bool, isIndex: bool }
           * <attr>: {id: <id>}
           *
           */

          JSONObject result = new JSONObject();

          JSONArray storageConfigurationGroups = new JSONArray();
          for (ERISLPConfiguration config : configsToUse) {
            JSONObject jsonGroup = new JSONObject();
            jsonGroup.put("newAPId", config.getAccessPath().getId());
            jsonGroup.put("newAPisDynamic", config.isDynamic());
            jsonGroup.put("isIndex", config.getAccessPath().getAddressingMode().equals("index"));
            JSONArray jsonArrayAttributes = new JSONArray();

            for (ERISAttribute erisAttribute : config.getAttributesAndKeys()) {
              JSONObject attribute = new JSONObject();
              attribute.put("id", erisAttribute.getId());
              jsonArrayAttributes.put(attribute);
            }
            jsonGroup.put("attributes", jsonArrayAttributes);
            storageConfigurationGroups.put(jsonGroup);

          }

          result.put("storageConfigurationGroups", storageConfigurationGroups);
          if (storageConfigurationGroups.isEmpty()) {
            logger.info("Not sending empty storage configuration group.");
          } else {
            // send result
            logger.info("           Sending result: {}", result);
            try {
              StringEntity params = new StringEntity(result.toString());

              URI uri = new URIBuilder()
                  .setHost(ERIS_URL)
                  .setScheme("http")
                  .setPath("swctrl/lp/" + livingPartition.getContainer().getId() + "/" + livingPartition.getNodeId() + "/" + livingPartition.getLocalId())
                  .setParameter("id", sessionID)
                  .build();
              HttpPost post = new HttpPost(uri);
              post.setEntity(params);

              String serverResponse = writeResult(post);
              logger.info("        Server responded with: {}", serverResponse);

            } catch (UnsupportedEncodingException | URISyntaxException e) {
              logger.error("Getting living partition failed: {}", e.getMessage());
              throw new RuntimeException(e);
            }
          }
        }
      }
    }
  }

  private String writeResult(HttpPost httpPost) {
    try {
      logger.info("Responding with: {}", httpPost.toString());
      CloseableHttpClient client = HttpClients.createDefault();

      CloseableHttpResponse response = client.execute(httpPost);
      Scanner s = new Scanner(response.getEntity().getContent()).useDelimiter("\\A");
      String result = s.hasNext() ? s.next() : "";

      client.close();
      response.close();

      return result;

    } catch (UnsupportedOperationException | IOException e) {
      logger.error("Writing result failed: {}", e.getMessage());
      throw new RuntimeException(e);
    }
  }

  public ERISModel getErisModel() {
    return this.erisModel;
  }


}
