package de.tudresden.inf.st.mquat.eris.coupling;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.json.JSONArray;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class ERISLPConfiguration {
  private static final Logger logger = LogManager.getLogger(ERISLPConfiguration.class);
  private ERISAccessPath accessPath;
  private List<ERISAttribute> attributes;
  private List<ERISAttribute> keys;
  private ERISLivingPartition livingPartition;
  private boolean dynamic;
  private Double estimatedCost = null;

  public ERISLPConfiguration(JSONObject jsonObject, ERISLivingPartition livingPartition) {

    this.livingPartition = livingPartition;

    dynamic = jsonObject.getBoolean("accessPathIsDynamic");

    // we do not know if this is the direct, indirect, or index access path
    // but we can assume that there is only one index access path.
    // dynamic -> indirect
    // not dynamic -> index OR direct
    if (dynamic) {
      accessPath = livingPartition.getContainer().getModel().getAccessPath(jsonObject.getInt("accessPathId"), "indirect");
    } else {
      accessPath = livingPartition.getContainer().getModel().getAccessPath(jsonObject.getInt("accessPathId"), "direct");
      if (accessPath == null) {
        accessPath = livingPartition.getContainer().getModel().getAccessPath(jsonObject.getInt("accessPathId"), "index");
      }
    }

    JSONArray lpAttributes = jsonObject.getJSONArray("attributes");
    attributes = new ArrayList<>();
    keys = new ArrayList<>();
    for (int i = 0; i < lpAttributes.length(); i++) {
      attributes.add(livingPartition.getContainer().getAttribute(lpAttributes.getJSONObject(i).getInt("id")));
    }
    JSONArray lpKeys = jsonObject.getJSONArray("keys");
    for (int i = 0; i < lpKeys.length(); i++) {
      attributes.add(livingPartition.getContainer().getAttribute(lpKeys.getJSONObject(i).getInt("id")));
    }

    // TODO what about the default access path?
  }

  public ERISLPConfiguration(ERISAccessPath accessPath, List<ERISAttribute> attributes, List<ERISAttribute> keys, ERISLivingPartition livingPartition, double estimatedCost) {
    this.accessPath = accessPath;
    this.attributes = attributes;
    this.keys = keys;
    this.livingPartition = livingPartition;
    this.estimatedCost = estimatedCost;
  }

  public ERISAccessPath getAccessPath() {
    return accessPath;
  }

  /**
   * This method is deprecated because almost always, both attributes AND keys are required
   *
   * @return
   */
  @Deprecated
  public List<ERISAttribute> getAttributes() {
    return attributes;
  }

  /**
   * This method is deprecated because almost always, both attributes AND keys are required
   *
   * @return
   */
  @Deprecated
  public List<ERISAttribute> getKeys() {
    return keys;
  }

  public List<ERISAttribute> getAttributesAndKeys() {
    List<ERISAttribute> all = new ArrayList<>();
    all.addAll(keys);
    all.addAll(attributes);
    return all;
  }

  public boolean isDynamic() {
    return dynamic;
  }

  public ERISLivingPartition getLivingPartition() {
    return livingPartition;
  }

  public Optional<Double> getEstimatedCost() {
    return Optional.ofNullable(estimatedCost);
  }

  public void setEstimatedCost(double estimatedCost) {
    this.estimatedCost = estimatedCost;
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    result.append("{[ ");
    for (ERISAttribute attribute : attributes) {
      result.append(attribute).append(" ");
    }
    result.append("],[ ");
    for (ERISAttribute key : keys) {
      result.append(key).append(" ");
    }
    result.append("] -> ").append(this.accessPath).append("}");

    return result.toString();
  }
}
