package de.tudresden.inf.st.mquat.solving;

import beaver.Parser;
import de.tudresden.inf.st.mquat.generator.ScenarioDescription;
import de.tudresden.inf.st.mquat.generator.ScenarioGenerator;
import de.tudresden.inf.st.mquat.jastadd.model.*;
import de.tudresden.inf.st.mquat.utils.ParserUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.rules.ErrorCollector;

import java.io.IOException;
import java.util.Iterator;
import java.util.concurrent.TimeUnit;

import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public abstract class HandwrittenTestSuite {
  private static Logger logger;
  @Rule
  public ErrorCollector collector = new ErrorCollector();
  private Solver solver;

  @BeforeClass
  public static void setupClass() {
    logger = LogManager.getLogger(HandwrittenTestSuite.class);
  }

  @Before
  public void setupSolverForTest() {
    this.solver = getSolver();
    solver.setTimeout(30, TimeUnit.SECONDS);
  }

  /**
   * Create and return the solver to use in this test suite
   *
   * @return a solver instance
   */
  protected abstract Solver getSolver();

  private Tuple<Root, Solution> loadAndSolve(String filename) throws IOException, Parser.Exception, SolvingException {
    Root model = ParserUtils.load(filename, HandwrittenTestSuite.class);
    Solution solution = solver.solve(model);
    if (logger.isDebugEnabled()) {
      MquatString out = solution.print(new MquatWriteSettings(" "));
      logger.debug("Solution:\n{}", out);
    }
    return new Tuple<>(model, solution);
  }

  private Assignment assertAssignment(Tuple<Root, Solution> modelAndSolution,
                                      int request, String impl, String resource) {
    Root model = modelAndSolution.getFirstElement();
    Assignment expectedAssignment = new Assignment();
    expectedAssignment.setRequest(model.getRequest(request));
    expectedAssignment.setImplementation(model.resolveImplementation(impl).get());
    expectedAssignment.setResourceMapping(new ResourceMapping(expectedAssignment.getImplementation().getResourceRequirement().getInstance(0), model.resolveResource(resource).get(), new List<>()));
    // check if assignment matches (partly) one listed in the solution
    Iterator<Assignment> assignmentIterator = modelAndSolution.getSecondElement().assignmentIterator();
    while (assignmentIterator.hasNext()) {
      Assignment actualAssignment = assignmentIterator.next();
      if (matches(actualAssignment, expectedAssignment)) {
        Assert.assertNotNull(actualAssignment);
        return actualAssignment;
      }
    }
    Assert.fail(String.format("Did not find match of assignment: %s on %s for request%s",
        impl, resource, request));
    throw new AssertionError();
  }

  private void assertComponentRequirement(Assignment requiringAssignment,
                                          String instanceName, Assignment expectedProvidingAssignment) {
    Instance instance = requiringAssignment.getImplementation().resolveInstance(instanceName).get();
    Assignment actualProvidingAssignment = requiringAssignment.mappedAssignment(instance);
    assertEquals(String.format("Not matching assignment for %s", instanceName),
        expectedProvidingAssignment, actualProvidingAssignment);
  }

  /**
   * Check if request, configuration and resource of the given assignments are equal
   *
   * @param actualAssignment   assignment in the computed solution
   * @param expectedAssignment expected assignment defined in the test case
   * @return <code>true</code> if both match, <code>false</code> otherwise
   */
  private boolean matches(Assignment actualAssignment, Assignment expectedAssignment) {
    return actualAssignment.getRequest().equals(expectedAssignment.getRequest()) &&
        actualAssignment.getImplementation().equals(expectedAssignment.getImplementation()) &&
        actualAssignment.getResource().equals(expectedAssignment.getResource());
  }

  private void assertValidSolution(Tuple<Root, Solution> modelAndSolution) {
//    Assert.assertTrue("Solution is not valid", modelAndSolution.getSecondElement().isValid());
    if (modelAndSolution.getSecondElement().getNumAssignment() == 0) {
      logger.warn("Solution is empty");
    }
    collector.checkThat("Solution is not valid", true,
        equalTo(modelAndSolution.getSecondElement().isValid()));

    logger.info("completeness: " + modelAndSolution.getSecondElement().evaluateSoftwareCompleteness());
    logger.info("sw-validity: " + modelAndSolution.getSecondElement().evaluateSoftwareValidity());
    logger.info("validity: " + modelAndSolution.getSecondElement().evaluateValidity());
    logger.info("objective: " + modelAndSolution.getSecondElement().computeObjective());
  }

  @Test
  public void test_01() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_01.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertComponentRequirement(config_0i0m0, "other", config_1i0m0);
  }

  @Test
  public void test_02() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_02.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertComponentRequirement(config_0i0m0, "other", config_1i0m0);
  }

  @Test
  public void test_03() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_03.txt");
    assertValidSolution(modelAndSolution);
    Assignment r0config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment r0config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    assertAssignment(modelAndSolution, 1, "config_1i0m0", "r2");
    assertComponentRequirement(r0config_0i0m0, "other", r0config_1i0m0);
  }

  @Test
  public void test_04() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_04.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r1");
    Assignment config_2i0m0 = assertAssignment(modelAndSolution, 0, "config_2i0m0", "r2");
    assertComponentRequirement(config_0i0m0, "alpha", config_1i0m0);
    assertComponentRequirement(config_0i0m0, "beta", config_2i0m0);
  }

  @Test
  public void test_05() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_05.txt");
    assertValidSolution(modelAndSolution);
    assertTrue(modelAndSolution.getSecondElement().isValid());
    Assignment configA = assertAssignment(modelAndSolution, 0, "configA0", "r0");
    Assignment configB = assertAssignment(modelAndSolution, 0, "configB0", "r1");
    Assignment configC = assertAssignment(modelAndSolution, 0, "configC0", "r4");
    Assignment configD = assertAssignment(modelAndSolution, 0, "configD0", "r3");
    Assignment configE = assertAssignment(modelAndSolution, 0, "configE0", "r2");
    Assignment configF = assertAssignment(modelAndSolution, 0, "configF0", "r5");
    Assignment configG = assertAssignment(modelAndSolution, 0, "configG0", "r6");
    assertComponentRequirement(configA, "beta", configB);
    assertComponentRequirement(configA, "epsilon", configE);
    assertComponentRequirement(configB, "chi", configC);
    assertComponentRequirement(configB, "delta", configD);
    assertComponentRequirement(configE, "phi", configF);
    assertComponentRequirement(configE, "gamma", configG);
    assertEquals("Objective does not fit", 21885.66,
        modelAndSolution.getSecondElement().computeObjective(), 0.01);
  }

  @Test
  public void test_06() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_06.txt");
    assertValidSolution(modelAndSolution);
    assertTrue(modelAndSolution.getSecondElement().isValid());
  }

  @Test
  public void test_paper() throws SolvingException {
    ScenarioDescription scenarioDescription = new ScenarioDescription(1, 2, 0, 0, 0, 2, 2, 1.1, 1, 1, 9);
    ScenarioGenerator generator = new ScenarioGenerator(scenarioDescription);
    Root model = generator.generate();

    Solution solution = solver.solve(model);
    if (logger.isDebugEnabled()) {
      logger.info("Model:\n{}",model.print(new MquatWriteSettings("  ")));
      logger.info("Solution:\n{}",solution.print(new MquatWriteSettings("  ")));
    }

    assertValidSolution(new Tuple<>(model, solution));
    assertTrue(solution.isValid());
  }

  @Test
  public void test_multi_01() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_multi_01.txt");
    assertValidSolution(modelAndSolution);
    Assignment config_0i0m0 = assertAssignment(modelAndSolution, 0, "config_0i0m0", "r0");
    Assignment config_1i0m0 = assertAssignment(modelAndSolution, 0, "config_1i0m0", "r0");
    assertComponentRequirement(config_0i0m0,"other", config_1i0m0);
  }

  @Test
  public void test_multi_eris_01() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_eris_01.txt");
    assertValidSolution(modelAndSolution);
    Assignment a0 = assertAssignment(modelAndSolution, 0, "ap4", "r0");
    Assignment a1 = assertAssignment(modelAndSolution, 1, "ap2", "r0");
  }

  @Test
  public void test_multi_eris_02() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_eris_02.txt");
    assertValidSolution(modelAndSolution);
  }

  @Test
  public void test_multi_eris_04() throws IOException, Parser.Exception, SolvingException {
    Tuple<Root, Solution> modelAndSolution = loadAndSolve("test_eris_04.txt");
    assertValidSolution(modelAndSolution);
  }
}
